(* ::Package:: *)

(************************************************************************)
(* This file was generated automatically by the Mathematica front end.  *)
(* It contains Initialization cells from a Notebook file, which         *)
(* typically will have the same name as this file except ending in      *)
(* ".nb" instead of ".m".                                               *)
(*                                                                      *)
(* This file is intended to be loaded into the Mathematica kernel using *)
(* the package loading commands Get or Needs.  Doing so is equivalent   *)
(* to using the Evaluate Initialization Cells menu command in the front *)
(* end.                                                                 *)
(*                                                                      *)
(* DO NOT EDIT THIS FILE.  This entire file is regenerated              *)
(* automatically each time the parent Notebook file is saved in the     *)
(* Mathematica front end.  Any changes you make to this file will be    *)
(* overwritten.                                                         *)
(************************************************************************)



(**************************************************************************

    This file is part of the LGSBloch package for Mathematica
    Copyright (C) 2009-2017 
    Simon M. Rochester <simon@rochesterscientific.com>
    Ron Holzl\[ODoubleDot]hner <rholzloe[at]eso.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)



(* ::Input::Initialization:: *)
BeginPackage["LGSBloch`LGSBeams`",{"AtomicDensityMatrix`","LGSBloch`"}]


(* ::Input::Initialization:: *)
Get["AtomicDensityMatrix`"]


(* ::Input::Initialization:: *)
Unprotect[Evaluate[$Context<>"*"]]


(* ::Input::Initialization:: *)
BeamRepositorySamples::usage="BeamRepositorySamples computes the beam repository samples." ;


(* ::Input::Initialization:: *)
BeamRepository::usage="BeamRepository computes a repository of return flux values. " ;


(* ::Input::Initialization:: *)
InterBeamMeso::usage="InterBeamMeso integrates beam return, and computes flux and beam diameter based on the 2nd moment." 


(* ::Input::Initialization:: *)
Sce::usage="Compute beam figure of merit \!\(\*FormBox[SubscriptBox[\(s\), \(ce\)],
TraditionalForm]\). Note that the following options may be changed from their values when BeamRepository[] was called, unless the peak intensity exceeds that of the interpolating function:
	PLaunch: Launched power [W]
	Zenith: Zenith angle [rad]   (note that MagneticZenith and MagneticAzimuth cannot be changed!)
	HNa: Altitude of sodium centroid above sea level [m]
	Htele: Altitude of launch telescope above sea level [m]
	DLT: Launch telescope aperture [m]
	\[Mu]LT: Ratio of DLT to \!\(\*FormBox[\(1/\*SuperscriptBox[\(e\), \(2\)]\),
TraditionalForm]\) lauch beam intensity diameter (not radius!), see Yura, Eq.(2.7)
	MsqrLT:  Effective \!\(\*FormBox[SuperscriptBox[\(M\), \(2\)],
TraditionalForm]\) beam quality factor out of the launch telescope, also taking into account the effect of speckles
	LaserSpotMesoFWHM: Mesospheric FWHM spot size [m]. If this option is set to a nonzero value, DLT and \[Mu]LT are not used to compute the mesospheric irradiance.

Output values:
	sceOut: Merit parameter \!\(\*FormBox[SubscriptBox[\(s\), \(ce\)],
TraditionalForm]\)\!\(\*
StyleBox[\" \",\nFontSlant->\"Italic\"]\)[ph/s/W/(atom/m^2)]
	GroundInt: Integrated flux on the ground [ph/s/atom]
	FluxMesoInt: Integrated flux in the mesosphere [ph*m^2/s/sr/atom]
	FWHMSpeckle: Instantaneous laser spot size in the mesosphere (central speckle) [m]
	FWHM2mom: Return light spot size based on second moment [m]
	IMesoMax: Peak mesospheric laser light intensity [W/m^2]" ;


(* ::Input::Initialization:: *)
AbsCrossSection::usage="Absorption cross section and cycle time

Input values: As for PsiMeso[]

Output values:
sig: absorption cross section [m^2]
ps: Result of PhiSec (computed for 200 velocity classes)  [ph/s/sr/(W/m^2)]
W: number of spontaneous excitations per time [ph/s/atom], inverse of cycle time (neglecting stimulated emission)

Example:
AbsCrossSection[{IMeso\[Rule]50.0,\[CapitalDelta]fLaser\[Rule]0.0*\!\(\*SuperscriptBox[\(10\), \(9\)]\),RepumpFraction\[Rule]0.12,BG\[Rule]0.0,LightPolarization\[Rule]1}]yields {1.1424135229284697`*^-15,402.5300545224092`,169414.22073090656`}, hence a cross-section of 1.14 times the low-irradiance cross section (1.0 x 10^(-11) cm^2), psi=402 ph/s/sr/(W/m^2), and the cycle time is 1/169414.22073090656`*10^6 us = 5.90 us, hence very close to the Larmor time in Paranal (5.8us)." ;


(* ::Input::Initialization:: *)
Begin["`Private`"]


(* ::Input::Initialization:: *)
w2f=Sqrt[2.0*Log[2]];(*Conversion factor from waist radius to FWHM of a Gaussian beam*)


(* ::Input::Initialization:: *)
ImaxWA[P_,FWHM_]:=Module[{A,w,Imax},
w=FWHM/w2f;
A=\[Pi] w^2;
Imax=2.0 P/A;
{Imax,w,A}
];


(* ::Input::Initialization:: *)
dSpeckle[Dap_,wLaunch_,\[Lambda]_,L_,Msqr_:1.0]:=4.0*Sqrt[Log[2]]/\[Pi]*\[Lambda]*L*Msqr/Dap*Sqrt[1+(Dap/2/wLaunch)^2]


(* ::Input::Initialization:: *)
(* MATLAB's linspace() routine for convenience *)
linspace[a_,b_,n_:100]:=Range[a,b,(b-a)/(n-1)];


(* ::Input::Initialization:: *)
Options[BeamRepositorySamples]=Join[{DefaultParameters:> $DefaultLGSParameters},{IMesoMax->0.0,IMesoInitial->0.35,MaxStepRatio->20,IMesoFactor->1.0,NSamples->6,LaserSpotMesoFWHM->0}]


(* ::Input::Initialization:: *)
BeamRepositorySamples[params_,opts:OptionsPattern[]]:=Module[{airmass,stepsNorm,fn,IMesoMax1,IMeso1Loc,FWHMMeso1,wLaunch,dist,IMesoAll,params1},
params1=Join[params,OptionValue[DefaultParameters]];
IMesoMax1=OptionValue[IMesoMax];
FWHMMeso1=OptionValue[LaserSpotMesoFWHM];
airmass = Sec[Zenith]/.params1;
If[IMesoMax1<=0.0,
If[FWHMMeso1<=0.0,
wLaunch=0.5DLT/\[Mu]LT/.params1;
(* distance to guide star in line of sight *)
dist=(HNa-Htele)*airmass/.params1;
(* Instantaneous FWHM spot size in mesosphere *)
FWHMMeso1=dSpeckle@@({DLT,wLaunch,LightWavelength,dist,MsqrLT}/.params1); 
];
IMesoMax1=ImaxWA[Ta^airmass*PLaunch/.params1,FWHMMeso1][[1]]; (* max intensity in mesosphere *)
];
If[OptionValue[NSamples]>1,
stepsNorm =linspace[1,OptionValue[MaxStepRatio],OptionValue[NSamples]-1]; (* compute vector of normalized steps *)
(* make sure iMeso1 won't become too large *)
IMeso1Loc=Min[OptionValue[IMesoInitial],IMesoMax1/10.0/OptionValue[NSamples]];
fn=(OptionValue[IMesoFactor]*IMesoMax1-IMeso1Loc)/Total[stepsNorm];  (* Normalization factor *)
IMesoAll=IMeso1Loc+fn*Join[{0},Accumulate[stepsNorm]];    (* Intensity vector *)
,(* ELSE *)
IMesoAll={OptionValue[IMesoInitial]};
];
{IMesoAll,IMesoMax1,FWHMMeso1}
] ;


(* ::Input::Initialization:: *)
Options[BeamRepository]=Join[Options[BeamRepositorySamples],{RefineResult->{100},InitVGParams->{{1,2},{1,3}},InitVG->Automatic,PlotVelocityGroups->False}];


(* ::Input::Initialization:: *)
BeamRepository[syst_,params_,opts:OptionsPattern[]]:=Module[{fluxAll,brs,brsopts,nfopts,IMesoMax1,FWHMMeso1},
SetSharedVariable[brs,nfopts];
brsopts=FilterRules[{{opts},Options[BeamRepository]},Options[BeamRepositorySamples]];
{brs,IMesoMax1,FWHMMeso1}=BeamRepositorySamples[params,brsopts];
nfopts=FilterRules[{{FluxReport->{},ReturnAllSteps->False},{opts},Options[BeamRepository]},Options[PsiMeso]];
fluxAll=Table[{brs[[k]],brs[[k]]PsiMeso[syst,Join[{IMeso->brs[[k]]},params],nfopts]},{k,1,Length[brs]}];
fluxAll=Join[{{0.0,0.0}},fluxAll];  (* add zero intensity point *)
{Interpolation[fluxAll,InterpolationOrder->2,If[$VersionNumber>=7.,Method->"Spline",{}]],fluxAll,IMesoMax1,FWHMMeso1}
];


(* ::Input::Initialization:: *)
Options[InterBeamMeso]=Options[BeamRepositorySamples];


(* ::Input::Initialization:: *)
InterBeamMeso[params_,PsiVector_,opts:OptionsPattern[]]:=Module[{w,IMesoMax1,r,FluxMesoInt,r2,airmass,FWHM2mom,GaussIntens,wLaunch,dist,FWHMMeso1,params1,Psi0,I0,c,fitFunction,fitRules,errTotRel,useSplines},
params1=Join[params,OptionValue[DefaultParameters]];
IMesoMax1=OptionValue[IMesoMax];
airmass = Sec[Zenith]/.params1;
FWHMMeso1=OptionValue[LaserSpotMesoFWHM];
If[IMesoMax1<=0,
If[FWHMMeso1<=0,
wLaunch=0.5DLT/\[Mu]LT/.params1;
(* distance to guide star in line of sight *)
dist=(HNa-Htele)*airmass/.params1;
(* Instantaneous FWHM spot size in mesosphere *)
FWHMMeso1=dSpeckle@@({DLT,wLaunch,LightWavelength,dist,MsqrLT}/.params1); 
];
(* Intensity and instantaneous FWHM spot size in mesosphere *)
{IMesoMax1,w}=ImaxWA[Ta^airmass*PLaunch/.params1,FWHMMeso1][[1;;2]];
,(* ELSE *)
w=If[OptionValue[LaserSpotMesoFWHM]<=0,
Sqrt[2.0/\[Pi] Ta^airmass PLaunch/IMesoMax1/.params1],FWHMMeso1/w2f];
FWHMMeso1=0.0;
];
(* Check that we're staying within the range of the simulated irradiance values *)
InterBeamMeso::IMaxOverflow= "Maximum intensity exceeds that of the supplied simulation data!";
If[IMesoMax1 > PsiVector[[-1,1]],Message[InterBeamMeso::IMaxOverflow];];
InterBeamMeso::PsiVector0= "PsiVector must start with the values {0,0}!";
If[PsiVector[[1]]!={0,0},Message[InterBeamMeso::PsiVector0];Abort[],{}];
If[verbosityLevel>3,
Print["--> InterBeamMeso[]: IMesoMax1: ",IMesoMax1,", w: ",w,", PLaunch: ",PLaunch/.params1];];
(* Assume Gaussian irradiance distribution in mesosphere. The Re[] is to avoid Less::nord errors in the numerical integration below *)
GaussIntens[r_]:=IMesoMax1*Exp[Re[-2.0(r/w)^2]];
(* Use splines instead of a logarithmic fit function if \[CapitalPsi](I) shows a roll-off *)
useSplines=If[PsiVector[[-1,2]]<Max[PsiVector[[All,2]]],True,False];
If[!useSplines,
(* Fit function template \[CapitalPsi](I)=Subscript[\[CapitalPsi], 0] ln[1+(I/Subscript[I, 0])^c]
Abs[] is to protect from imaginary function values. Using a fit function is much more reliable than employing Interpolation[], even with InterpolationOrder\[Rule]2 *)
fitFunction[x_]:=Psi0 Log[1.0+(x/Abs[I0] )^Abs[c]];
(* Fit model function to PsiVector. Supply initial parameter values so that the value of Psi at the last simulation point equals Psi0*ln(2) *)
fitRules=FindFit[PsiVector,fitFunction[x],{{Psi0, PsiVector[[-1,2]]/Log[2.0]},{I0, PsiVector[[-1,1]]},{c,1.0}},x,AccuracyGoal->6,MaxIterations->300];
(* Total relative error (1st moment) *)
errTotRel=Total[Abs[PsiVector[[All,2]]-fitFunction[PsiVector[[All,1]]]/.fitRules]]/Total[PsiVector[[All,2]]];
(* Fit can fail, e.g. in case of roll-over of \[CapitalPsi][I]. Use spline interpolation in this case. *)
InterBeamMeso::BadFit="Large error in return flux fit of `1`%! Trying spline interpolation.";
If[errTotRel>0.02,
Message[InterBeamMeso::BadFit,N[100.0*errTotRel,4]];
useSplines=True;
];
, (* ELSE *)
fitRules={};
];(*If[!useSplines]*)
If[useSplines,
fitFunction=Interpolation[PsiVector,InterpolationOrder->2,Method->"Spline"];
errTotRel=Total[Abs[PsiVector[[All,2]]-fitFunction[PsiVector[[All,1]]]/.fitRules]]/Total[PsiVector[[All,2]]];
];
(* Turn off this error message *)
Off[InverseFunction::ifun];
(* Carry out numerical integral, exploiting the rotational symmetry of a Gaussian. FluxMesoInt was denoted by Subscript[F, m] in Holzloehner et al.,A&A 510,A20 (2010), Eq.(7) *)
FluxMesoInt=NIntegrate[2.0\[Pi] r  fitFunction[GaussIntens[r]]/.fitRules,{r,0.0,3.0w},AccuracyGoal->5];
(* 2nd moment computation (weight by r^2) *)
r2=NIntegrate[2.0\[Pi]  r^3  fitFunction[GaussIntens[r]]/.fitRules,{r,0.0,3.0w},AccuracyGoal->5];
On[InverseFunction::ifun];
(* 2nd moment. In the absence of any saturation (hence a/b-->Infinity and c\[Equal]1), this equals the FWHM beam diameter. *)
FWHM2mom=w2f*Sqrt[2.0r2/FluxMesoInt];
 {FluxMesoInt,FWHMMeso1, FWHM2mom,IMesoMax1,fitFunction,fitRules,errTotRel}
] ;



(* ::Input::Initialization:: *)
Options[Sce]=Options[InterBeamMeso]


(* ::Input::Initialization:: *)
Sce[params_,PsiVector_,opts:OptionsPattern[]]:=Module[{sceOut,FluxMesoInt,FWHM2mom,IMesoMax1,airmass,FluxGroundInt,FWHMMeso1,params1,fitFunction,fitRules,errTotRel},
{FluxMesoInt,FWHMMeso1,FWHM2mom,IMesoMax1,fitFunction,fitRules,errTotRel}=
InterBeamMeso[params,PsiVector,opts];
params1=Join[params,OptionValue[DefaultParameters]];
airmass = Sec[Zenith]/.params1;
FluxGroundInt=Ta^airmass*FluxMesoInt/(HNa-Htele)^2/airmass/.params1;
sceOut=FluxMesoInt/PLaunch/Ta^airmass/.params1;
{sceOut,FluxGroundInt,FluxMesoInt,FWHMMeso1,FWHM2mom,IMesoMax1,fitFunction,fitRules,errTotRel}
] ;


(* ::Input::Initialization:: *)
Options[AbsCrossSection]=FilterRules[Options[PsiMeso],Except[{ReturnAllSteps,FluxReport}]]


(* ::Input::Initialization:: *)
AbsCrossSection[syst_,params_,opts:OptionsPattern[]]:=Module[{ps,exStates,fb2,W,constc,consth,nu,sig,nfopts,params1},
params1=Join[params,OptionValue[DefaultParameters]];
nfopts=FilterRules[{{FluxReport->{DMSolution,BlochMatrix},ReturnAllSteps->False},opts,Options[AbsCrossSection]},Options[PsiMeso]];
ps=PsiMeso[syst,params,nfopts];
fb2=(BlochMatrix/.ps[[2]])[[3]]; 
exStates=Flatten[Position[fb2,x_/;x>0]]; (* Find indices of radiating states *)
fb2[[exStates]]=6.154149127341655`*10^7; (* Neglect antenna factors (set to 1.0) *)W=Re[fb2.(DMSolution/.ps[[2]])];
constc=2.99792*10^8;
consth=6.626069*10^(-34);
nu=consth*constc/LightWavelength/.params1; (*  Compute frequency of D2b line *)
sig=nu*W/IMeso/.params1;(* absorption cross-section in m *)
{sig,ps[[1]],W}
] ;


(* ::Input::Initialization:: *)
End[]


(* ::Input::Initialization:: *)
Protect[Evaluate[$Context<>"*"]]


(* ::Input::Initialization:: *)
EndPackage[]
