(* ::Package:: *)

(************************************************************************)
(* This file was generated automatically by the Mathematica front end.  *)
(* It contains Initialization cells from a Notebook file, which         *)
(* typically will have the same name as this file except ending in      *)
(* ".nb" instead of ".m".                                               *)
(*                                                                      *)
(* This file is intended to be loaded into the Mathematica kernel using *)
(* the package loading commands Get or Needs.  Doing so is equivalent   *)
(* to using the Evaluate Initialization Cells menu command in the front *)
(* end.                                                                 *)
(*                                                                      *)
(* DO NOT EDIT THIS FILE.  This entire file is regenerated              *)
(* automatically each time the parent Notebook file is saved in the     *)
(* Mathematica front end.  Any changes you make to this file will be    *)
(* overwritten.                                                         *)
(************************************************************************)



(**************************************************************************

    This file is part of the LGSBloch package for Mathematica
    Copyright (C) 2009-2017 
    Simon M. Rochester <simon@rochesterscientific.com>
    Ron Holzl\[ODoubleDot]hner <rholzloe[at]eso.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)



(* ::Input::Initialization:: *)
BeginPackage["LGSBloch`StepwiseConstantSolver`",{"AtomicDensityMatrix`", "LGSBloch`"}]


(* ::Input::Initialization:: *)
Get["AtomicDensityMatrix`"]


(* ::Input::Initialization:: *)
Unprotect[Evaluate[$Context<>"*"]]


(* ::Input::Initialization:: *)
phiv::usage="[w, err] = phiv( t, A, u, v, tol, m )
 phiv computes an approximation of w = exp(t*A)*v + t*phi(t*A)*u
 for a general matrix A using Krylov subspace projection techniques.
 Here, phi(z) = (exp(z)-1)/z and w is the solution of the 
 nonhomogeneous linear ODE problem w' = Aw + u, w(0) = v.
 It does not compute the matrix functions in isolation but instead,
 it computes directly the action of these functions on the 
 operand vectors. This way of doing so allows for addressing large
 sparse problems. The matrix under consideration interacts only
 via matrix-vector products (matrix-free method)." 


(* ::Input::Initialization:: *)
VerbosityLevel::usage="" 


(* ::Input::Initialization:: *)
StepPulseFlux::usage="Solve the time-dependent problem for periodic irradiation pattern and integrate flux return over the last period. Uses phiv[] to solve the inhomogeneous problem (d/dt)\[Rho] = A\[Rho]+\!\(\*
StyleBox[\"b\",\nFontSlant->\"Italic\"]\) for piecewise constant coefficients.

Input:

\!\(\*
StyleBox[\"syst\",\nFontSlant->\"Italic\"]\): Bloch equation system
	
\!\(\*
StyleBox[\"params\",\nFontSlant->\"Italic\"]\): Physical simulation parameters for PsiMeso[]

\!\(\*
StyleBox[\"nperiodsIn\",\nFontSlant->\"Italic\"]\): maximum number of periods to simulate
	
\!\(\*
StyleBox[\"complexE\",\nFontSlant->\"Italic\"]\): Vector, subdividing the pulse into n subspans, scaling the irradiance in each subspan so that I\[LeftDoubleBracket]j\[RightDoubleBracket] == IMeso |complexE\[LeftDoubleBracket]j,2\[RightDoubleBracket]|^2, and Arg[complexE\[LeftDoubleBracket]j,2\[RightDoubleBracket]] is the optical phase at step j. The entries complexE\[LeftDoubleBracket]j,1\[RightDoubleBracket] contain the corresponding time steps relative to the beginning of the simulated period. Hence, complexE\[LeftDoubleBracket]-1,1\[RightDoubleBracket] equals the duration of the simulated period. The time steps do not have to be constant. Note that all time steps, including complexE\[LeftDoubleBracket]1,1\[RightDoubleBracket], must be larger than 0 and strictly increasing.
	
\!\(\*
StyleBox[\"brightPeriodSkip\",\nFontSlant->\"Italic\"]\): Concentrate the light to every BrightPeriodSkip'th period (e.g., if brightPeriodSkip = 2, run every other period at twice the irradiance as when setting brightPeriodSkip = 1). Set brightPeriodSkip = 1 to make every period the same.
	
\!\(\*
StyleBox[\"\[CapitalDelta]fLaserFunction\",\nFontSlant->\"Italic\"]\): \[CapitalDelta]fLaserFunction[] sets the laser central frequency offset as a function of time
	
\!\(\*
StyleBox[\"fname\",\nFontSlant->\"Italic\"]\): File name to save the results structure \"res\" in, using Export[fname,res]
	
\!\(\*
StyleBox[\"opts\",\nFontSlant->\"Italic\"]\): option set for PsiMeso[]

Output:

Global`Res = {psiAvg, psiAvgTotal, psiSpanAll, psiAvgHist, Iavg, popgr8All, popex16All, {PhiVErrors \[Rule] errAll, PsiIavg \[Rule] psiIavg1, GroundPopulation \[Rule] popgr, ExcitedPopulation \[Rule] popex, FinalGroundPops \[Rule] groundPops, FinalExcitedPops \[Rule] excitedPops, ExcitedPops \[Rule] excitedPopsAll, LastDM \[Rule] {w, vg}, RFParamsOut \[Rule] {rfparams1, params1, opts}}}

\!\(\*
StyleBox[\"psiAvg\",\nFontSlant->\"Italic\"]\): Specific return flux averaged over the last period [ph/s/sr/atom/(W/m^2)]

\!\(\*
StyleBox[\"psiAvgTotal\",\nFontSlant->\"Italic\"]\): psiAvg summed over all periods [ph/s/sr/atom/(W/m^2)]

\!\(\*
StyleBox[\"psiSpanAll\",\nFontSlant->\"Italic\"]\): vector of pairs {tn, return flux(tn)}, return flux [ph/s/sr/atom/(W/m^2)]

\!\(\*
StyleBox[\"psiAvgHist\",\nFontSlant->\"Italic\"]\): vector of psiAvg values for each period

\!\(\*
StyleBox[\"Iavg\",\nFontSlant->\"Italic\"]\): Irradiance, averaged over a simulation period [W/m^2]

\!\(\*
StyleBox[\"popgr8All\",\nFontSlant->\"Italic\"]\): Array of dimension Nx8, where N equals the length of psiSpanAll and corresponds to the time steps given by psiSpanAll\[LeftDoubleBracket]All,1\[RightDoubleBracket]. The values of opgr8All yield the 8 ground state populations of the velocity class that is closest to +5 MHz, where in many systems a peak of the polarization is observed (\"busy velocity class\"), normalized by the total ground+excited state population in this velocity class. For darkness in equilibrium we thus have popgr8All\[LeftDoubleBracket]All,All\[RightDoubleBracket]==1/8. The states are ordered {(F=2,m=2), (F=2,m=1)..., (F=2,m=-2), (F=1,m=1), (F=1,m=0), (F=2,m=-1)} (with LightPolarization->+1, the (F=2,m=2) is pumped most strongly) popex16All: same as popgr8All, but for the 16 excited states. Sequence starts with (F=3,m=3).

\!\(\*
StyleBox[\"PhiVErrors\",\nFontSlant->\"Italic\"]\): vector of numerical errors returned by phiv[], corresponding to psiSpanAll psiIavg: value of psi for average irradiance

\!\(\*
StyleBox[\"popgr\",\nFontSlant->\"Italic\"]\): ground state population

\!\(\*
StyleBox[\"popex\",\nFontSlant->\"Italic\"]\): Excited state population

\!\(\*
StyleBox[\"w\",\nFontSlant->\"Italic\"]\): final solution vector (b)

The routine computes the average return flux psiAvg defined as

   psiAvg = 1/(trep * Iavg) \!\(\*SubsuperscriptBox[\(\[Integral]\), \(t0\), \(t0\\\  + \\\ trep\)]\)Psi(t) dt, 

	where Psi(t) is the instantaneous non-normalized return flux in ph/s/sr/atom, t0 is the begin of the last period, and Iavg = Ipulse*duty, where Ipulse is the pulse irradiance.

The iteration starts from thermal equilibrium (InitialState \[Rule] LightOff) or from the steady-state solution at Ipulse (InitialState \[Rule] LightOn) and stops once the difference of psiAvg between two subsequent iterations drops below some threshold. In the verbose simulation output, \"peak cw psi\" is the steady-state result psi(I_peak), and \"flux from avg irradiance\" is similarly psi(I_avg). To achieve convergence, allow for a total simulation time of several times 16ns. At low irradiance, you may have to simulate several collision time scales." ;


(* ::Input::Initialization:: *)
KrylovDimension::usage="KrylovDimension is an option for StepPulseFlux that specifies the dimension of the Krylov subspace to be used in the matrix exponential calculation.";


(* ::Input::Initialization:: *)
Tolerance::usage="Tolerance is an option for StepPulseFlux that specifies the tolerance to use in phiv[].";


(* ::Input::Initialization:: *)
TerminationThresh::usage="TerminationThresh is an option for StepPulseFlux that specifies that the iteration shall be terminated once the absolute difference between two subsequent values of the averaged psi falls below TerminationThresh";


(* ::Input::Initialization:: *)
InitialState::usage="InitialState is an option for StepPulseFlux that specifies the state of the system at the beginning of the time-dependent calculation.

InitialState\[Rule]LightOn specifies the steady state with cw light.
InitialState\[Rule]LightOff specifies the steady state with no light.
InitialState\[Rule]{\!\(\*
StyleBox[\"sol\", \"TI\"]\),\!\(\*
StyleBox[\"vg\", \"TI\"]\)} specifies a state given by density-matrix solution vector \!\(\*
StyleBox[\"sol\", \"TI\"]\) corresponding to velocity group binning \!\(\*
StyleBox[\"vg\", \"TI\"]\).";


(* ::Input::Initialization:: *)
LightOn::usage="LightOn is a possible value for the InitialState option of StepPulseFlux that specifies that the initial state is the steady state with cw light.";


(* ::Input::Initialization:: *)
LightOff::usage="LightOff is a possible value for the InitialState option of StepPulseFlux that specifies that the initial state is the steady state with no light.";


(* ::Input::Initialization:: *)
VGMethod::usage="VGMethod is an option for StepPulseFlux that specifies the method used to account for the velocity group binning over multiple steps.

VGMethod->FixVG specifies that the velocity group binning found for the first calculation (using the nominal laser parameters) should be reused for all subsequent steps.
VGMethod->ConvertDM specifies that a new binning should be found for each step, and the density matrix solution vector found from the previous step should be rebinned to match the new velocity group binning.";


(* ::Input::Initialization:: *)
FixVG::usage="FixVG is a possible value for the VGMethod option for StepPulseFlux that specifies that the velocity group binning found for the first calculation (using the nominal laser parameters) should be reused for all subsequent steps.";


(* ::Input::Initialization:: *)
ConvertDM::usage="ConvertDM is a possible value for the VGMethod option for StepPulseFlux that specifies that a new binning should be found for each step, and the density matrix solution vector found from the previous step should be rebinned to match the new velocity group binning.";


(* ::Input::Initialization:: *)
VGfreezePeriod::usage=
"VGfreezePeriod is the period after starting at which the rebinning stops and the velocity bins remain constant. In particular when simulating pulse shapes with numerous samples, this option can significantly improve the computational speed.";


(* ::Input::Initialization:: *)
PhiVErrors::usage="";


(* ::Input::Initialization:: *)
PsiIavg::usage="";


(* ::Input::Initialization:: *)
GroundPopulation::usage="";


(* ::Input::Initialization:: *)
ExcitedPopulation::usage="";


(* ::Input::Initialization:: *)
FinalGroundPops::usage="";


(* ::Input::Initialization:: *)
FinalExcitedPops::usage="";


(* ::Input::Initialization:: *)
ExcitedPops::usage="";


(* ::Input::Initialization:: *)
LastDM::usage="";


(* ::Input::Initialization:: *)
RFParamsOut::usage="";


(* ::Input::Initialization:: *)
Begin["`Private`"]


(* ::Input::Initialization:: *)
If[$VersionNumber>=8.,Needs["CCompilerDriver`"]];


(* ::Input::Initialization:: *)
ccompiler=$VersionNumber>=8.&&CCompilers[]!={};


(* ::Input::Initialization:: *)
compileropts=If[ccompiler,{RuntimeOptions->"Speed",Parallelization->True, CompilationTarget->"C" },{}]


(* ::Input::Initialization:: *)
Options[phiv]={VerbosityLevel->2};


(* ::Input::Initialization:: *)
phiv[t_,A_,u_,v_,obsmat_:{},tspanMax_:Infinity,tol_:10.0^(-7),mIn_:-1,indVgMaxEx_:-1,OptionsPattern[]]:=Module[{n,anorm,mxrej,btol,gamma,delta,mb,tout,istep,tnew,tnow,serror,rndoff,k1,xm,beta,fact,s,sgn,w,tstep,Vt,H,p,i,j,avnorm,ireject,mx,errloc,p1,p2,err,m,h,obs,F,off1,time,verb,popgr8,popex16,grStates,exStates,wgr,wex,tot1,popgr8Weighted,wgrAll,wexAllTotal,popgr8WeightedLoc},
(*function (w,err,hump)=expv (t,A,v,tol,m)*)
n=Dimensions[A][[1]];
If[{n}!=Dimensions[u],Print["phiv[]: A must be conformable with v, Dimensions[A]\[LeftDoubleBracket]1\[RightDoubleBracket]: ", {n}, ", Dimensions[u]: ", Dimensions[u]];Abort[],{};,Print["u"];Abort[]];
If[{n}!=Dimensions[v],Print["phiv[]: A must be conformable with v, Dimensions[A]\[LeftDoubleBracket]1\[RightDoubleBracket]: ", {n}, ", Dimensions[v]: ", Dimensions[v]];Abort[],{};,Print["v"];Abort[]];
If[MatrixQ[A,NumericQ]==False ,Print["phiv[]: A is nonnumeric"];Abort[],{};,Print["A"];Abort[]];
If[MatrixQ[{u},NumericQ]==False ,Print["phiv[]: u is nonnumeric"];Abort[],{};,Print["A"];Abort[]];
If[MatrixQ[{v},NumericQ]==False ,Print["phiv[]: v is nonnumeric"];Abort[],{};,Print["A"];Abort[]];
If[MatrixQ[{obsmat},NumericQ]==False ,Print["phiv[]: resultmat is nonnumeric"];Abort[],{};,Print["A"];Abort[]];

verb=OptionValue[VerbosityLevel];

(* Positions of the ground state populations *)
grStates=Flatten[Position[u,x_/;x!=0]];
(* Positions of the excited state populations *)
exStates=Flatten[Position[obsmat,x_/;x>0]];
popgr8={};
popex16={};
popgr8Weighted={};

(*[n,n]=size (A);*)(*if nargin\[Equal]3,tol=1.0*10^(-7);
m=min (n,30);
end;
if nargin\[Equal]4,m=min (n,30);
end;*)
If[mIn==-1,m=Min[n,30];,m=mIn;];
anorm=Norm[A,Infinity];
mxrej=10;btol=10.0^(-7);
gamma=0.9;delta=1.2;
mb=m;tout=N[Abs[t]];
istep=0;tnew=0;
tnow=0;serror=0;
rndoff=anorm*$MachineEpsilon;
sgn=N[Sign[t]];
k1=3;
xm=1.0/m;
obs={}; (* Used to accumulate return flux samples *)

w=v;
While[tnow<tout,
time=Timing[
Vt=ConstantArray[0.0,{m+1,n}];(*V=zeros (n,m+1);*)
H=ConstantArray[0.0,{m+3,m+3}];(*H=zeros (m+3,m+3);*)
Vt[[1]]=A . w+u;
beta=Norm[Vt[[1]],2];
Vt[[1]]*=(1.0/beta);
If[istep==0,
fact=(((m+1.0)/E)^(m+1))*Sqrt[2.0*Pi*(m+1)];
tnew=(1.0/anorm)*((fact*tol)/(4*beta*anorm))^xm;
s=10.0^(Floor[Log10[tnew]]-1);
tnew=Ceiling[tnew/s]*s;
];
istep++;
tstep=Min[tout-tnow,tnew];
tstep=Min[tstep,tspanMax];

(* V\[LeftDoubleBracket]All,1\[RightDoubleBracket]=(1.0/beta)*w;    V (:,1)=(1/beta)*w;*)
For[j=1,j<=m,j++,(*for j=1:m*)
p=A.Vt[[j]];(*p=A*V (:,j);*)
For[i=1,i<=j,i++,(*for i=1:j*)
H[[i,j]]=Conjugate[Vt[[i]]].p;(*H (i,j)=V (:,i)'*p;*)
p-=H[[i,j]] * Vt[[i]];(*p=p-H (i,j)*V (:,i);*)
];(*for i*)
s=Norm[p,2];(*s=norm (p);*)
If[s<btol,(*if s<btol,*)
k1=0;
mb=j;
tstep=tout-tnow;
tstep=Min[tstep,tspanMax];
Break[](*break;*)
];(* if *)
H[[j+1,j]]=s;(*H (j+1,j)=s;*)
(*V (:,j+1)=(1/s)*p;*)
Vt[[j+1]]=(1.0/s)*p;
];(*for j;*)
H[[1,mb+1]]=1;

If[k1!=0,(*if k1~=0,*)
H[[m+1,m+2]]=1.0;
H[[m+2,m+3]]=1.0;
h=H[[m+1,m]];
H[[m+1,m]]=0.0;
(*avnorm=norm (A*V (:,m+1));*)
avnorm=Norm[A.Vt[[m+1]] ];
];(*if*)

ireject=0;
While[ireject<=mxrej,(*while ireject\[LessEqual]mxrej,*)
mx=mb+Max[1,k1];
(* Re[] added on 09Sep11 since MatrixExp[] sometimes returns complex numbers *)
F=Re[MatrixExp[ sgn*tstep* H[[1;;mx,1;;mx]] ]];(*F=expm (sgn*t_step*H (1:mx,1:mx));*)
If[k1==0,(*if k1\[Equal]0,*)
errloc=btol;
Break[];,(*break;*)
(* Else *)
F[[m+1,m+1]]=h*F[[m,m+2]];
F[[m+2,m+1]]=h*F[[m,m+3]];
p1=Abs[beta*F[[m+1,m+1]]];(*abs (beta*F (m+1,m+1));*)
p2=Abs[beta*F[[m+2,m+1]]*avnorm];(*abs (beta*F (m+2,m+1)*avnorm);*)

If[p1>10.0*p2,(*if phi1>10*phi2,*)
errloc=p2;
xm=1.0/m;,
(* Else *)
If[p1>p2,
errloc=(p1*p2)/(p1-p2);
xm=1.0/m;,
(*Else*)
errloc=p1;
xm=1.0/(m-1);
];
]; 
];(* If k\[Equal]1 *)
If[errloc<=delta*tstep*tol,(*if err_loc\[LessEqual]delta*t_step*tol,*)
Break[];,
(*Else*)
tstep=gamma*tstep*(tstep*tol/errloc)^xm;
tstep=Min[tstep,tspanMax];
s=10.0^(Floor[Log10[tstep]]-1);
tstep=Ceiling[tstep/s]*s;
If[ireject==mxrej,
Print["phiv[]: The requested tolerance is too high!"];
Abort[];
];
ireject++;
];
];(* While[ireject\[LessEqual]mxrej,...] *)

mx=mb+Max[0,k1-2];
w=Transpose[Vt[[1;;mx]]].(beta*F[[1;;mx,mb+1]]) + w;
(*w=V (:,1:mx)*(beta*F (1:mx,mb+1)) + w;*)

tnow+=tstep;(*t_now=t_now+t_step;*)
tnew=gamma*tstep*(tstep*tol/errloc)^xm;(*t_new=gamma*t_step*(t_step*tol/err_loc)^xm;*)
s=10.0^(Floor[Log10[tnew]]-1);(*s=10^(floor (log10 (t_new))-1);*)
tnew=Ceiling[tnew/s]*s;(*t_new=ceil (t_new/s)*s;*)
errloc=Max[errloc,rndoff];(*err_loc=max (err_loc,rndoff);*)
serror+=errloc;
];

If[istep<=4&&verb>3||verb>5,
Print["    Step ",istep,": tstep = ", 10.0^9*tstep, " ns", ", tnow = ", 10.0^9*tnow, " ns, (",time[[1]]," s)"];];
(* Code inserted by Ron Holzloehner: Compute the return flux at tnow *)
If[n==Max[Dimensions[obsmat]],
AppendTo[obs,{tnow,Re[obsmat.w]}];
];

If[indVgMaxEx>=1,
off1=8*(indVgMaxEx-1);
wgr=w[[grStates]][[(1+off1);;(8+off1)]];
off1=16*(indVgMaxEx-1);
wex=w[[exStates]][[(1+off1);;(16+off1)]];
(* Normalize by number of atoms in this velocity bin *)
tot1=N[Total[wgr]+Total[wex]];
popgr8=Join[popgr8,{wgr/tot1}];
popex16=Join[popex16,{wex/tot1}];
, (*else*)
popgr8=Join[popgr8,{ConstantArray[-1.0,8]}];
popex16=Join[popex16,{ConstantArray[-1.0,16]}];
];
(* Nx8 array, where N is the number of velocity classes  *)
wgrAll=Partition[w[[grStates]],8];
(* Vector of length N with the total excited populations (total of all 16 states) in each velocity class *)
wexAllTotal=Total[Partition[w[[exStates]],16],{2}];
tot1=Total[wexAllTotal];
If[tot1>0,
(* Vector of length 8 containing the ground state populations weighted by the excited state populations *)
popgr8WeightedLoc=Total[wgrAll*wexAllTotal,{1}]/tot1;
(* Normalize to total 1 *)
popgr8WeightedLoc/=Total[popgr8WeightedLoc];
popgr8Weighted=Join[popgr8Weighted,{popgr8WeightedLoc}];
, (* ELSE *)
popgr8Weighted=Join[popgr8Weighted,{ConstantArray[-1.0,8]}];
];

]; (* end of While[] *)
err=serror;
{w,err,obs,popgr8,popex16,popgr8Weighted}
]; (*End of Module[] phiv[] *)


(* ::Input::Initialization:: *)
(* Modified version of phiv[] with a compiled inner loop *)


(* ::Input::Initialization:: *)
Options[phivCompiled]={VerbosityLevel->2};
(* The sparse matrix A is passed to multiplyByA[] via the global variable Global`fullAmatG *)
phivCompiled[t_,u_,v_,obsmat_:{},tspanMax_:1.0*^200,tol_:1.0*^-7,mIn_:-1,indVgMaxEx_:-1,OptionsPattern[]]:=Module[{n,anorm,tout,istep,tnew,tnow,serror,rndoff,sgn,w,err,m,obs,off1,time,verb,popgr8,popex16,grStates,exStates,wgr,wex,tot1,popgr8Weighted,wgrAll,wexAllTotal,popgr8WeightedLoc,tstep,xm,k1,mb,out1,errloc,s},

(*function (w,err,hump)=expv (t,A,v,tol,m)*)
n=Dimensions[Global`fullAmatG][[1]];
If[{n}!=Dimensions[u],Print["phivCompiled[]: A must be conformable with v, Dimensions[A]\[LeftDoubleBracket]1\[RightDoubleBracket]: ", {n}, ", Dimensions[u]: ", Dimensions[u]];Abort[],{};,Print["u"];Abort[]];
If[{n}!=Dimensions[v],Print["phivCompiled[]: A must be conformable with v, Dimensions[A]\[LeftDoubleBracket]1\[RightDoubleBracket]: ", {n}, ", Dimensions[v]: ", Dimensions[v]];Abort[],{};,Print["v"];Abort[]];
(* Since the matrix A doesn't have to be passed, check here if the 2nd argument u is a matrix instead of a vector *)
If[MatrixQ[u,NumericQ]==True ,Print["phivCompiled[]: Second argument should be the vector u, not a matrix (different argument list than phiv[])!"];Abort[],{};];
If[MatrixQ[Global`fullAmatG,NumericQ]==False ,Print["phivCompiled[]: A is nonnumeric"];Abort[],{};];
If[MatrixQ[{u},NumericQ]==False ,Print["phivCompiled[]: u is nonnumeric"];Abort[];];
If[MatrixQ[{v},NumericQ]==False ,Print["phivCompiled[]: v is nonnumeric"];Abort[];];
If[MatrixQ[{obsmat},NumericQ]==False &&obsmat!={},Print["phivCompiled[]: resultmat is nonnumeric"];Abort[];];

verb=OptionValue[VerbosityLevel];

(* Positions of the ground state populations *)
grStates=Flatten[Position[u,x_/;x!=0]];
(* Positions of the excited state populations *)
exStates=Flatten[Position[obsmat,x_/;x>0]];
popgr8={};
popex16={};
popgr8Weighted={};

(*[n,n]=size (A);*)(*if nargin\[Equal]3,tol=1.0*10^(-7);
m=min (n,30);
end;
if nargin\[Equal]4,m=min (n,30);
end;*)
m=If[mIn==-1,Min[n,30],mIn];
anorm=Norm[Global`fullAmatG,Infinity];
tout=N[Abs[t]];
istep=0;tnew=0;
tnow=0;serror=0;
rndoff=anorm*$MachineEpsilon;
sgn=N[Sign[t]];
xm=1.0/m;
mb=m;
k1=3;
obs={}; (* Used to accumulate return flux samples *)

w=v;
(*  =============== Main Loop ===================  *)
While[tnow<tout,

time=Timing[
(* Due to restrictions of Compile[], the output is a single vector with the following elements: {w,istep,tstep,tnew,tnow,errloc,s,xm,mb}
All floating point variables are real, not complex  *)
out1=phivLoopCompiled[m,n,w,u,istep,tout,tnow,tnew,tspanMax,sgn,tol,rndoff,anorm,xm,mb,k1];
w=out1[[1;;Length[v]]];
istep=Round[out1[[-8]]];
tstep=out1[[-7]];tnew=out1[[-6]]; tnow=out1[[-5]];
errloc=out1[[-4]];s=out1[[-3]];
xm=out1[[-2]];mb=Round[out1[[-1]]];
serror+=errloc;
]; (* End of Timing[] *)

If[istep<=4&&verb>3||verb>5,
Print["    Step ",istep,": tstep = ", 10.0^9*tstep, " ns", ", tnow = ", 10.0^9*tnow, " ns, (",time[[1]]," s)"];];

(* Code inserted by Ron Holzloehner: Compute the return flux at tnow *)
If[n==Max[Dimensions[obsmat]],
AppendTo[obs,{tnow,Re[obsmat.w]}];
];

If[indVgMaxEx>=1,
off1=8*(indVgMaxEx-1);
wgr=w[[grStates]][[(1+off1);;(8+off1)]];
off1=16*(indVgMaxEx-1);
wex=w[[exStates]][[(1+off1);;(16+off1)]];
(* Normalize by number of atoms in this velocity bin *)
tot1=N[Total[wgr]+Total[wex]];
popgr8=Join[popgr8,{wgr/tot1}];
popex16=Join[popex16,{wex/tot1}];
, (*else*)
popgr8=Join[popgr8,{ConstantArray[-1.0,8]}];
popex16=Join[popex16,{ConstantArray[-1.0,16]}];
];
(* Nx8 array, where N is the number of velocity classes  *)
wgrAll=Partition[w[[grStates]],8];
(* Vector of length N with the total excited populations (total of all 16 states) in each velocity class *)
wexAllTotal=Total[Partition[w[[exStates]],16],{2}];
tot1=Total[wexAllTotal];
If[tot1>0,
(* Vector of length 8 containing the ground state populations weighted by the excited state populations *)
popgr8WeightedLoc=Total[wgrAll*wexAllTotal,{1}]/tot1;
(* Normalize to total 1 *)
popgr8WeightedLoc/=Total[popgr8WeightedLoc];
popgr8Weighted=Join[popgr8Weighted,{popgr8WeightedLoc}];
, (* ELSE *)
popgr8Weighted=Join[popgr8Weighted,{ConstantArray[-1.0,8]}];
];

]; (* end of While[] *)
err=serror;
{w,err,obs,popgr8,popex16,popgr8Weighted}
]; (*End of Module[]*)


(* ::Input::Initialization:: *)
(* Inner loop of Expokit's phiv routine, sped up with Compile[]. All floating point variables are of type real, not complex.
Ron Holzloehner, 12Sep2011
 *)


(* ::Input::Initialization:: *)
phivLoopCompiled=Compile[{{m,_Integer},{n,_Integer},{wIn,_Real,1},{u,_Real,1},{istepIn,_Integer},{tout,_Real},{tnowIn,_Real},{tnewIn,_Real},{tspanMax,_Real},{sgn,_Real},{tol,_Real},{rndoff,_Real},{anorm,_Real},{xmIn,_Real},{mbIn,_Integer},{k1In,_Integer}},Module[{mxrej=10,btol=1.0*^-7,gamma=0.9,delta=1.2,mb=mbIn,k1=k1In,xm=xmIn,beta=0.0,fact=0.0,s=0.0,tstep=0.0,Vt=ConstantArray[0.0,{m+1,n}],Vt1=multiplyByA[ wIn]+u,H=ConstantArray[0.0,{m+3,m+3}],p=u,i,j,avnorm=0.0,ireject=0,mx=0,errloc=0.0,p1=0.0,p2=0.0,h=0.0,F=ConstantArray[0.0,{2,2}],tnow=tnowIn,tnew=tnewIn,istep=istepIn,w=wIn,Hij=0.0,Vti=u},

(*
Vt=ConstantArray[0.0,{m+1,n}];(*V=zeros (n,m+1);*)
H=ConstantArray[0.0,{m+3,m+3}];(*H=zeros (m+3,m+3);*)
*)

(* A has dimensions  {n,Length[w]} *)
(* Vt1 = A.w+u;   <-- Now in parameter initialization *)
beta=Norm[Vt1,2];
(* V\[LeftDoubleBracket]All,1\[RightDoubleBracket]=(1.0/beta)*w;    V (:,1)=(1/beta)*w;*)
Vt1*=(1.0/beta);
Vt[[1]]=Vt1;
If[istep==0,
fact=(((m+1.0)/E)^(m+1))*Sqrt[2.0*Pi*(m+1)];
tnew=(1.0/anorm)*((fact*tol)/(4*beta*anorm))^xm;
s=10.0^(Floor[Log10[tnew]]-1);
tnew=Ceiling[tnew/s]*s;
];
istep++;
tstep=Min[tout-tnow,tnew];
tstep=Min[tstep,tspanMax];

For[j=1,j<=m,j++,(*for j=1:m*)
p=multiplyByA[Vt[[j]]];(*p=A*V (:,j);*)
(* The following loop eats up 80% or so of the total CPU time *)
For[i=1,i<=j,i++,(*for i=1:j*)
(* WARNING: The following line should use the Conjugate[] function. We omit it to speed up the code for type real only!!! *)
(* H\[LeftDoubleBracket]i,j\[RightDoubleBracket]=Conjugate[Vt\[LeftDoubleBracket]i\[RightDoubleBracket]].p;(*H (i,j)=V (:,i)'*p;*)*)
Vti=Vt[[i]];
(* H\[LeftDoubleBracket]i,j\[RightDoubleBracket]=Vti.p;*)
Hij=Vti.p;
p-=Hij * Vti;(*p=p-H (i,j)*V (:,i);*)
H[[i,j]]=Hij;
];(*for i*)
s=Norm[p,2];(*s=norm (p);*)
If[s<btol,(*if s<btol,*)
k1=0;
mb=j;
tstep=tout-tnow;
tstep=Min[tstep,tspanMax];
Break[](*break;*)
];(* if *)
H[[j+1,j]]=s;(*H (j+1,j)=s;*)
(*V (:,j+1)=(1/s)*p;*)
Vt[[j+1]]=(1.0/s)*p;
];(*for j;*)
H[[1,mb+1]]=1;

If[k1!=0,(*if k1~=0,*)
H[[m+1,m+2]]=1.0;
H[[m+2,m+3]]=1.0;
h=H[[m+1,m]];
H[[m+1,m]]=0.0;
(*avnorm=norm (A*V (:,m+1));*)
avnorm=Norm[multiplyByA[Vt[[m+1]]] ];
];(*if*)

ireject=0;
While[ireject<=mxrej,(*while ireject\[LessEqual]mxrej,*)
mx=mb+Max[1,k1];
(* If Range[1,mx] is replaced by 1;;mx, we get a compilation error *)
(* Re[] added on 09Sep11 since MatrixExp[] sometimes returns complex numbers *)
(* Runtime of MatrixExp[] was checked with Timing[] and was negligible *)
F=Re[MatrixExp[ (sgn*tstep)* H[[Range[1,mx],1;;mx]] ]];
(*F=expm (sgn*t_step*H (1:mx,1:mx));*)
If[k1==0,(*if k1\[Equal]0,*)
errloc=btol;
Break[];,(*break;*)
(* Else *)
F[[m+1,m+1]]=h*F[[m,m+2]];
F[[m+2,m+1]]=h*F[[m,m+3]];
p1=Abs[beta*F[[m+1,m+1]]];(*abs (beta*F (m+1,m+1));*)
p2=Abs[(beta*avnorm)*F[[m+2,m+1]]];(*abs (beta*F (m+2,m+1)*avnorm);*)

If[p1>10.0*p2,(*if phi1>10*phi2,*)
errloc=p2;
xm=1.0/m;,
(* Else *)
If[p1>p2,
errloc=(p1*p2)/(p1-p2);
xm=1.0/m;,
(*Else*)
errloc=p1;
xm=1.0/(m-1);
];
]; 
];(* If k\[Equal]1 *)
If[errloc<=delta*tstep*tol,(*if err_loc\[LessEqual]delta*t_step*tol,*)
Break[];,
(*Else*)
tstep=gamma*tstep*(tstep*tol/errloc)^xm;
tstep=Min[tstep,tspanMax];
s=10.0^(Floor[Log10[tstep]]-1);
tstep=Ceiling[tstep/s]*s;
If[ireject==mxrej,
Print["phivLoopCompiled[]: The requested tolerance is too high!"];
Abort[];
];
ireject++;
];
];(* While[ireject\[LessEqual]mxrej,...] *)

mx=mb+Max[0,k1-2];
w=Transpose[Vt[[1;;mx]]].(beta*F[[1;;mx,mb+1]]) + w;
(*w=V (:,1:mx)*(beta*F (1:mx,mb+1)) + w;*)

tnow+=tstep;(*t_now=t_now+t_step;*)
tnew=gamma*tstep*(tstep*tol/errloc)^xm;(*t_new=gamma*t_step*(t_step*tol/err_loc)^xm;*)
s=10.0^(Floor[Log10[tnew]]-1);(*s=10^(floor (log10 (t_new))-1);*)
tnew=Ceiling[tnew/s]*s;(*t_new=ceil (t_new/s)*s;*)
errloc=Max[errloc,rndoff];(*err_loc=max (err_loc,rndoff);*)
(*  This command has been moved up to the calling routine
serror+=errloc;*)

(* Combine everything in one vector since Compile[] cannot return irregular arrays *)
Join[w,N[{istep,tstep,tnew,tnow,errloc,s,xm,mb}]]
] (* End of Module *)
,Evaluate[Sequence@@compileropts]
];(* End of Compile *)


(* ::Input::Initialization:: *)
(* Multiply a vector v by the global matrix variable Global`fullAmatG (which may be a sparse matrix, in which case Compile[] couldn't handle the multiplication) *)
multiplyByA[v_]:=(Global`fullAmatG.v)


(* ::Input::Initialization:: *)
(* Stripped version of phiv[] to work with with Mathematica's MatrixExp[] function (see StepPulseFlux[]) *)


(* ::Input::Initialization:: *)
Options[phivMatrixExp]={VerbosityLevel->2};
phivMatrixExp[t_,w_,grStates_,exStates_,obsmat_,indVgMaxEx_:-1,OptionsPattern[]]:=Module[{verb,popgr8,popex16,wgr,wex,tot1,popgr8Weighted,wgrAll,wexAllTotal,err,off1,obs,popgr8WeightedLoc},

If[MatrixQ[{w},NumericQ]==False ,Print["phivMatrixExp[]: w is nonnumeric"];Abort[];];
If[MatrixQ[{obsmat},NumericQ]==False &&obsmat!={},Print["phivCompiled[]: resultmat is nonnumeric"];Abort[];];

verb=OptionValue[VerbosityLevel];


popgr8={};
popex16={};
popgr8Weighted={};
obs={}; (* Used to accumulate return flux samples *)

(* Code inserted by Ron Holzloehner: Compute the return flux at tnow *)
AppendTo[obs,{t,Re[obsmat.w]}];

If[indVgMaxEx>=1,
off1=8*(indVgMaxEx-1);
wgr=w[[grStates]][[(1+off1);;(8+off1)]];
off1=16*(indVgMaxEx-1);
wex=w[[exStates]][[(1+off1);;(16+off1)]];
(* Normalize by number of atoms in this velocity bin *)
tot1=N[Total[wgr]+Total[wex]];
popgr8=Join[popgr8,{wgr/tot1}];
popex16=Join[popex16,{wex/tot1}];
, (*else*)
popgr8=Join[popgr8,{ConstantArray[-1.0,8]}];
popex16=Join[popex16,{ConstantArray[-1.0,16]}];
];
(* Nx8 array, where N is the number of velocity classes  *)
wgrAll=Partition[w[[grStates]],8];
(* Vector of length N with the total excited populations (total of all 16 states) in each velocity class *)
wexAllTotal=Total[Partition[w[[exStates]],16],{2}];
tot1=Total[wexAllTotal];
If[tot1>0,
(* Vector of length 8 containing the ground state populations weighted by the excited state populations *)
popgr8WeightedLoc=Total[wgrAll*wexAllTotal,{1}]/tot1;
(* Normalize to total 1 *)
popgr8WeightedLoc/=Total[popgr8WeightedLoc];
popgr8Weighted=Join[popgr8Weighted,{popgr8WeightedLoc}];
, (* ELSE *)
popgr8Weighted=Join[popgr8Weighted,{ConstantArray[-1.0,8]}];
];

err=0.0;
{err,obs,popgr8,popex16,popgr8Weighted}
]; (*End of Module[] phivMatrixExp *)


(* ::Input::Initialization:: *)
convertDM[olddm_,oldvg_,newvg_]:=Module[{oldbins,newbins,dm,convertlist,leftpos,leftfrac,middlepos,middlefrac,rightpos,rightfrac,containedpos,containedfrac,inf,leftovers},
oldbins=oldvg[[All,3]]/.-\[Infinity]->-inf/.\[Infinity]->inf;
newbins=newvg[[All,3]]/.-\[Infinity]->-inf/.\[Infinity]->inf;
dm=Partition[olddm,Length[olddm]/Length[oldbins]];
convertlist=Table[
leftpos=Position[oldbins,{x1_,x2_}/;x1<bin[[1]]<x2<bin[[2]]];
leftfrac=(#[[2]]-bin[[1]])/(#[[2]]-#[[1]])&/@Extract[oldbins,leftpos];
middlepos=Position[oldbins,{x1_,x2_}/;bin[[1]]<= x1<=x2<=bin[[2]]];
middlefrac=1&/@Extract[oldbins,middlepos];
rightpos=Position[oldbins,{x1_,x2_}/;bin[[1]]<x1<bin[[2]]<x2];
rightfrac=(bin[[2]]-#[[1]])/(#[[2]]-#[[1]])&/@Extract[oldbins,rightpos];
containedpos=Position[oldbins,{x1_,x2_}/;(x1<=bin[[1]]<=bin[[2]]<x2)||(x1<bin[[1]]<=bin[[2]]<=x2)];
containedfrac=(bin[[2]]-bin[[1]])/(#[[2]]-#[[1]])&/@Extract[oldbins,containedpos];
Flatten[{Transpose@{leftfrac,leftpos},Transpose@{middlefrac,middlepos},Transpose@{rightfrac,rightpos},Transpose@{containedfrac,containedpos}},1],
{bin,newbins}];
leftovers=Chop[Table[{1-Total@Cases[Flatten[convertlist,1],{x_,{i}}->x],{i}},{i,Length[oldbins]}],2$MachineEpsilon];
convertlist[[1]]=DeleteCases[Take[leftovers,Floor[Length[oldbins]/2]],{0,_}];
convertlist[[-1]]=DeleteCases[Drop[leftovers,Floor[Length[oldbins]/2]],{0,_}];
Flatten[Sum[conv[[1]]Extract[dm,conv[[2]]],{conv,#}]&/@convertlist]
]


(* ::Input::Initialization:: *)
(*  Convert velocity groups for use within StepPulseFlux[] *)


(* ::Input::Initialization:: *)
Options[makeVg]=Join[{DefaultParameters:>$DefaultLGSParameters},Options[ReturnFlux]];
SetOptions[makeVg,RefineResult->{100}]


(* ::Input::Initialization:: *)
makeVg[speclist_,atdata1_,params_,opts:OptionsPattern[]]:=Module[{dw,nog,allfreqs,vg,freqs,spec,bendfunc,wf,nb},
nb=NDither/.params;
dw=DopplerWidth/.params;
nog=OptionValue[RefineResult]/.False|None->{};
freqs=ResonanceFrequencies/.atdata1;
allfreqs=Union@Flatten@Table[{freqs-DetuningA[i]/.params,freqs-DetuningB[i]/.params},{i,nb}];
spec=Interpolation[speclist,InterpolationOrder->1];
bendfunc=Interpolation[bendList[speclist],InterpolationOrder->1];
wf=nWeightFunction[spec,bendfunc,{5,5,5},{x,-3dw,3dw}];
vg=VelocityGroups2[wf,{x,-3dw,3dw},nog[[-1]],dw,5 RecoilShift/.params,allfreqs,ShowPlot->OptionValue[PlotVelocityGroups]]
(**)
]


(* ::Input::Initialization:: *)
filterPops[sol_,vec_,vg_,returnSublevels_:False]:=Module[{pops},
pops=getPops[sol,vec];
If[Mod[Length[pops]/Length[vg],1]!=0||Length[pops]/Length[vg]<1,Print["filterPops[]: Length[pops] is not a multiple of Length[vg], Aborting!"];Abort[];];
Transpose@If[returnSublevels,
{vg[[All,1]]/(2.0\[Pi] 10^9),-(2.0\[Pi] 10^9)(Partition[pops,Length[pops]/Length[vg]])/Subtract@@@vg[[All,3]]}
,
{vg[[All,1]]/(2.0\[Pi] 10^9),-(2.0\[Pi] 10^9)(Total/@Partition[pops,Length[pops]/Length[vg]])/Subtract@@@vg[[All,3]]}
]
];


(* ::Input::Initialization:: *)
getPops[sol_,vec_]:=sol[[Flatten[Position[vec,x_/;x!=0]]]];


(* ::Input::Initialization:: *)
Options[StepPulseFlux]=FilterRules[{Options[PsiMeso],{VGMethod->ConvertDM,InitialState->LightOff,KrylovDimension->Automatic,VerbosityLevel->2,Tolerance->10.0^(-7),TerminationThresh->0.25*10^(-2),VGfreezePeriod->12}},Except[{FluxReport,ReturnAllSteps}]];


(* ::Input::Initialization:: *)
StepPulseFlux[syst_,params_,nperiodsIn_,complexE_,brightPeriodSkip_,\[CapitalDelta]fLaserFunction_,fname_,opts:OptionsPattern[]]:=Module[{trep,tNow,i,fullAmat,w,err,errAll,psiSpan,psiSpanAll,tend,fns,hh,psiAvg,psiAvgTotal,tns,tspanMaxFactor,tol,m,Iavg,psiIavg1,popgr,popex,fullbvecOn,psiAvgLast,psiAvgHist,j,errLoc,psiSpanLoc,IMesoNow,relativeI,phase,tDiffs,psi,inst,lastvg,vg,nfopts,params1,sol,fullfluorvecOn,fullbvecOff,fullfluorvecOff,bvec,fvec,groundPops,excitedPops,excitedPopsAll,excitedPopsHelp,speclist,NDitherLoc,rfparams1,indVgMaxEx,popgr8,popex16,popgr8All,popex16All,\[CapitalDelta]fLaserNow,tu1,tu2,tu12,tu13,tuLast,terminationCriterion,fwriteCriterion,help1,tu11,tPrep,popgr8Weighted,popgr8WeightedAll,iVGfreezeLoc,fullAmatLI,A110L,grStates,exStates,nperiodsFile,nperiods,factorEnhance},

(* Argument checking *)
If[complexE[[1,1]]<=0,Print["First time step must be larger than zero. Aborting."];Abort[];];
(* Period *)
trep=complexE[[-1,1]];
If[trep<=0.0||trep>0.1,Print["trep: ", trep,"s, which may not be what you want. Aborting."];Abort[];];
(* Massage options*)
inst=OptionValue[InitialState];
(* Tolerance for phiv[]*)
tol=OptionValue[Tolerance];
(* Krylov subspace dimension (-1 means default:30)*)
m=OptionValue[KrylovDimension]/.Automatic->-1;
(* Solver parameters *)
nfopts=FilterRules[{ReturnAllSteps->False,opts,Options[StepPulseFlux]},Options[PsiMeso]];
(* LGS Parameters for PsiMeso[] *)
params1=Join[params,OptionValue[DefaultParameters]];
(* Normalize the relative irradiances by the peak irradiance in time domain *)
relativeI=Abs[complexE[[All,2]]]^2/Max[Abs[complexE[[All,2]]]^2];
phase=Arg[complexE[[All,2]]];
(* Factor from seconds to nanoseconds*)
fns=1.0*10^9;
(* Minimum number of time steps per constant intensity span*)
tspanMaxFactor=11.0;
(* Speed-up computation of fullAmat starting with this period *)
iVGfreezeLoc=If[OptionValue[VGMethod]===FixVG,1,OptionValue[VGfreezePeriod]];

(* time sample steps *)
tDiffs=Differences[complexE[[All,1]]];
PrependTo[tDiffs,complexE[[1,1]]];
(* Average irradiance (use piecewise constant irradiance, exactly like input irradiance) *)
Iavg=(relativeI.tDiffs)/trep*IMeso/.params1;

(* Run steady state calculation at unperturbed intensity and frequency *)
{w,vg}=If[Head[inst]===List,If[Length[inst]!=2,Print["StepPulseFlux[]: Option InitialState must either be a keyword or a list of length 2! Aborting..."];Abort[]];
inst,{Indeterminate,Automatic}];
\[CapitalDelta]fLaserNow=\[CapitalDelta]fLaserFunction[0.5*tDiffs[[1]]];
psi=PsiMeso[syst,Join[{\[CapitalDelta]fLaser->\[CapitalDelta]fLaserNow},params],FluxReport->{BlochMatrix,DMSolution,VelocityGroups,RFParams},InitVG->vg,nfopts];
If[Head[w]===List&&Length[w]>0,(* Both w and vg have been supplied *)
{{fullAmat,fullbvecOn,fullfluorvecOn},rfparams1}={BlochMatrix,RFParams}/.psi[[2]];
,
If[Head[vg]===List&&Length[vg]>0,
inst=inst[[1]]; (* w was not initialized to a constant: in this case, expect LightOn or LightOff in inst\[LeftDoubleBracket]1\[RightDoubleBracket]  *)
{{fullAmat,fullbvecOn,fullfluorvecOn},w,rfparams1}={BlochMatrix,DMSolution,RFParams}/.psi[[2]];
,
{{fullAmat,fullbvecOn,fullfluorvecOn},w,vg,rfparams1}={BlochMatrix,DMSolution,VelocityGroups,RFParams}/.psi[[2]];
];
];

(* indVgMaxEx is the index of the velocity group nearest to +5MHz, where we usually find a peak in the spectrum. We pass this index to phiv[] to extract population information of this "busy" velocity group. *)
(*help1=Abs[vg\[LeftDoubleBracket]All,1\[RightDoubleBracket]-0.5*NaturalWidth[2]/.Last[syst]];
indVgMaxEx=Position[help1,Min[help1]]\[LeftDoubleBracket]1,1\[RightDoubleBracket];*)
help1=Total[Partition[getPops[w,fullfluorvecOn],16],{-1}]; (* List of total exitation per velocity bin *)
indVgMaxEx=Position[help1,Max[help1]][[1,1]];
If[OptionValue[VerbosityLevel]>3,Print["   Bin with max excitation: ",indVgMaxEx, ", memory size of fullAmat: ",ByteCount[fullAmat]/1024.0^2," MB"];
];

(*Initialize with steady-state vector for I\[Equal]IMeso (the system Aw\[Equal]-b was solved and use the formula fullbvec\[Equal]-\[Gamma]b)*)
If[inst===LightOn,
bvec=fullbvecOn;fvec=fullfluorvecOn;
, (* ELSE *)
If[inst===LightOff,
(*Run steady-state calculation at close to zero intensity*)psi=PsiMeso[syst,Join[{IMeso->$MachineEpsilon,\[CapitalDelta]fLaser->\[CapitalDelta]fLaserNow},params],RefineResult->{},FluxReport->{BlochMatrix,DMSolution},InitVG->vg,nfopts];
{{fullAmat,fullbvecOff,fullfluorvecOff},w}={BlochMatrix,DMSolution}/.psi[[2]];
bvec=fullbvecOff;fvec=fullfluorvecOff;
If[Dimensions[fullbvecOn]!= Dimensions[fullbvecOff],
Print["StepPulseFlux[]: Dimensions[fullbvecOn]: ", Dimensions[fullbvecOn], " != Dimensions[fullbvecOff]: ", Dimensions[fullbvecOff],", aborting"];Abort[];];
,(* ELSE *)
Print["StepPulseFlux[]: Found inst = ",inst," but should equal one of LightOn or LightOff, aborting..."];Abort[];
];(*  If[inst===LightOff] *)
];(*  If[inst===LightOn] *)

If[Dimensions[w]!= Dimensions[fullbvecOn],
Print["StepReturnFlux[]: Dimensions[w]: ", Dimensions[w], " != Dimensions[fullbvecOn]: ", Dimensions[fullbvecOn],", aborting"];Abort[];];

(* Compute return flux from *average* irradiance *)
psiIavg1=PsiMeso[syst,Join[{IMeso->Iavg,\[CapitalDelta]fLaser->\[CapitalDelta]fLaserNow},params],RefineResult->{},FluxReport->{},InitVG->vg,nfopts];
If[OptionValue[VerbosityLevel]>0,
Print["I: ",IMeso/.params1," W/m^2, Iavg: ",Iavg,", frep: ",10^-6/trep " MHz, peak cw psi: ",psi[[1]]," ph/s/sr/atom/(W/m^2), psi from Iavg (cw): ",psiIavg1,", total excited state population at I: ",Total[getPops[w,fullfluorvecOn]]];
];


psiSpanAll={{0.0,psi[[1]]}};
errAll={};
tend=0.0;psiAvgLast=0.0;psiAvgTotal=0.0;
(*Ground and excited state populations*)
popgr={Partition[getPops[w,fullbvecOn],8]};
popex={{0.0,Total[getPops[w,fvec]]}};
(*Here we assume the "Off" inital state.Worth generalizing?*)
popgr8All={ConstantArray[1.0/8,8]};
popex16All={ConstantArray[0.0,16]};
popgr8WeightedAll={ConstantArray[1.0/8,8]};
psiAvgHist={};
excitedPopsAll={};
tuLast=TimeUsed[];
lastvg=vg;
tPrep=0.0;
(* Positions of the ground state populations *)
grStates=Flatten[Position[fullbvecOn,x_/;x!=0]];
(* Positions of the excited state populations *)
exStates=Flatten[Position[fullfluorvecOn,x_/;x>0]];
If[OptionValue@VerbosityLevel >0,Print["Dimensions[grStates]: ",Dimensions[grStates],", Dimensions[exStates]: ",Dimensions[exStates]]
];



(* ============================== MAIN LOOP BEGIN =======================  *)
nperiods=nperiodsIn;
For[i=1,i<=nperiods,i++,
(* CPU time spent in this session so far *)
tu1=TimeUsed[];

If[FileExistsQ["NPERIODS.dat"], (* Check to see if there's a file NPERIODS.dat and if it contains a single integer larger tan zero, if so, terminate iteration at the latest after that pulse *)
nperiodsFile=ToExpression[Import["NPERIODS.dat","Text"]];
If[IntegerQ[nperiodsFile]&&nperiodsFile>=1&&nperiodsFile!=nperiods,
nperiods=Max[i,nperiodsFile];
Print["\n********** StepPulseFlux[]: Read file NPERIODS.dat, will terminate after pulse ",nperiods];
];
];

If[OptionValue[VerbosityLevel]>2,
(* Current central frequency offset, possibly changed due to chirping *)
\[CapitalDelta]fLaserNow=\[CapitalDelta]fLaserFunction[tend+0.5*tDiffs[[1]]];
Print["  Laser period ",i,"/",nperiods," of duration ",fns*trep," ns, DeltafLaserNow: ",10.0^-6*\[CapitalDelta]fLaserNow," MHz"];
];
err={};
psiSpan={};

(*
If[i\[Equal]iVGfreezeLoc,
(*Added by S.R.: create light-independent part of Bloch matrix*)
psi=PsiMeso[SparseReplace[syst,{RabiFrequencyA[1]\[Rule]0,RabiFrequencyB[1]\[Rule]0,DetuningA[1]\[Rule]0,DetuningB[1]\[Rule]0,LightPhase[1]\[Rule]0}],params,RefineResult\[Rule]{},FluxReport\[Rule]{BlochMatrix},InitVG\[Rule]vg,SkipLinearSolve\[Rule]True,nfopts];
{fullAmatLI,fullbvecOn,fullfluorvecOn} = BlochMatrix/.psi\[LeftDoubleBracket]2\[RightDoubleBracket];
(*Added by S.R.: create light-only part of one block of Bloch matrix*)
A110L=SparseArray@@(Expand[{ArrayRules[syst\[LeftDoubleBracket]2\[RightDoubleBracket]-SparseReplace[syst\[LeftDoubleBracket]2\[RightDoubleBracket],{RabiFrequencyA[1]\[Rule]0,RabiFrequencyB[1]\[Rule]0,DetuningA[1]\[Rule]0,DetuningB[1]\[Rule]0,LightPhase[1]\[Rule]0}]],Dimensions[syst\[LeftDoubleBracket]2\[RightDoubleBracket]]}]/.({x_Integer,y_Integer}\[Rule]0)\[Rule]Sequence[]);
If[OptionValue[VerbosityLevel]>2&&i>1,Print["  --> Freezing velocity groups (rebinning stopped)"];];
];
*)
(*  Bright period: Integrate over sub time spans *)
For[j=1,j<=Length[relativeI],j++,
(* time since beginning of current period *)
tNow=complexE[[j,1]];
(** Concentrate the irradiance to every BrightPeriodSkip'th period Set brightPeriodSkip=1 to make all periods even **)
factorEnhance=If[Mod[i-1,brightPeriodSkip]==0,brightPeriodSkip,0];
(* Current irradiance. A value of zero leads to a 1/0 division *)
IMesoNow=Max[factorEnhance*IMeso*relativeI[[j]]/.params1,$MachineEpsilon];
(* Current central frequency offset, possibly changed due to chirping *)
(* \[CapitalDelta]fLaserNow=(tend+0.5*tDiffs\[LeftDoubleBracket]j\[RightDoubleBracket])*chirpRate+\[CapitalDelta]fLaser/.params1;*)
\[CapitalDelta]fLaserNow=\[CapitalDelta]fLaserFunction[tend+0.5*tDiffs[[j]]];

(* Just assemble the Bloch matrices; don't actually solve the equations *)
tu11=TimeUsed[];
(*Added by S.R.: for fixed velocity groups, just add light-dependent part to precalculated light-independent part.
Commented out by R.H. on 17Nov11 in order to accommodate MatrixExp[], see below.
*)
(*
If[i\[GreaterEqual]iVGfreezeLoc,
Global`fullAmatG=fullAmatLI+BuildLightDependentPart[syst\[LeftDoubleBracket]1\[RightDoubleBracket],A110L,syst\[LeftDoubleBracket]-1\[RightDoubleBracket],Join[{IMeso\[Rule]IMesoNow,Phase\[Rule]phase\[LeftDoubleBracket]j\[RightDoubleBracket],\[CapitalDelta]fLaser\[Rule]\[CapitalDelta]fLaserNow},params],vg]
, (* Otherwise, create entire matrix *)

psi=PsiMeso[syst,Join[{IMeso\[Rule]IMesoNow,Phase\[Rule]phase\[LeftDoubleBracket]j\[RightDoubleBracket],\[CapitalDelta]fLaser\[Rule]\[CapitalDelta]fLaserNow},params],RefineResult\[Rule]{},FluxReport\[Rule]{BlochMatrix},InitVG\[Rule]vg,SkipLinearSolve\[Rule]True,nfopts];
{Global`fullAmatG,fullbvecOn,fullfluorvecOn} = BlochMatrix/.psi\[LeftDoubleBracket]2\[RightDoubleBracket];
];
*)

(* We do solve the full system here to update the variable "sol" (option DMSolution), used in MatrixExp[] below  *)
psi=PsiMeso[syst,Join[{IMeso->IMesoNow,Phase->phase[[j]],\[CapitalDelta]fLaser->\[CapitalDelta]fLaserNow},params],RefineResult->{},FluxReport->{BlochMatrix,DMSolution},InitVG->vg,SkipLinearSolve->False,nfopts];
{{fullAmat,fullbvecOn,fullfluorvecOn},sol} = {BlochMatrix,DMSolution}/.psi[[2]];
tu12=TimeUsed[];
tPrep+=tu12-tu11;

(*
Export["vg_speclist_"<>ToString[Round[IMesoNow]]<>"Wm2_"<>ToString[i]<>"_"<>ToString[j]<>".wdx",{vg,speclist,syst,rfparams1,nfopts}];
*)

(*[w,err,...]=phivCompiled[t_,u_,v_,obsmat_:{},tspanMax_:1.0*^200,tol_:1.0*^-7,mIn_:-1,indVgMaxEx_:-1,OptionsPattern[]] solves the ODE w' = Aw+b, w(0) = v *)
(*
Global`fullAmatG=fullAmat;
{w,errLoc,psiSpanLoc,popgr8,popex16,popgr8Weighted}=phivCompiled[tDiffs\[LeftDoubleBracket]j\[RightDoubleBracket],fullbvecOn,w,fullfluorvecOn,trep/tspanMaxFactor,tol,m,indVgMaxEx,VerbosityLevel\[Rule]OptionValue[VerbosityLevel]-1];
*)

(* Solver for constant light field based on Mathematica's routine MatrixExp[m,v], faster than phiv[] and phivCompiled[] in many cases  *)
w=Re[MatrixExp[tDiffs[[j]]*fullAmat,(w+sol)]-sol];
{errLoc,psiSpanLoc,popgr8,popex16,popgr8Weighted}=phivMatrixExp[tDiffs[[j]],w,grStates,exStates,fullfluorvecOn,indVgMaxEx,VerbosityLevel->OptionValue[VerbosityLevel]-1];
tu13=TimeUsed[];

If[OptionValue[VerbosityLevel]>4||(OptionValue[VerbosityLevel]>3&&i==1),Print["    Subspan ",j,"/",Length[relativeI]," of ",fns*tDiffs[[j]]," ns, I: ",IMesoNow, " W/m^2, t: ",1.0*^6(tend+0.5*tDiffs[[j]])," us, DfLaserNow: ",1.0*^-6\[CapitalDelta]fLaserNow," MHz, Phase: ",phase[[j]],", CPU time {PsiMeso,MatrixExp}: ",{tu12-tu11,tu13-tu12}," s"];
];

AppendTo[err,errLoc];
psiSpanLoc[[All,1]]+=tend;(* Advance time axis (tend is the time from beginning of first pulse) *)
psiSpan=Join[psiSpan,psiSpanLoc];
tend+=tDiffs[[j]];
popex=Join[popex,{{fns*tend,Total[getPops[w,fullfluorvecOn]]}}];
popgr8All=Join[popgr8All,popgr8];
popex16All=Join[popex16All,popex16];
popgr8WeightedAll=Join[popgr8WeightedAll,popgr8Weighted];
];(*For[j]*)
AppendTo[popgr,Partition[getPops[w,fullbvecOn],8]];

(* Scale time axis to ns to avoid precision loss in Integrate[] *)
psiSpan[[All,1]]*=fns;
(* Divide return flux samples by average intensity *)
psiSpan[[All,2]]/=Iavg;
(* Integrate Flux during each period. Prepend last value of previous span in order to properly begin the integral. Changed to InterpolationOrder\[Rule]1 to correctly integrate long-span data points  *)
hh=Interpolation[Join[{psiSpanAll[[-1,All]]},psiSpan],InterpolationOrder->1];
(* Must use NIntegrate[] with Method\[Rule]"Spline" in Interpolation[], not Integrate[] *)
psiAvg=NIntegrate[hh[tns],{tns,hh[[1,1,1]],hh[[1,1,2]]}]/(fns*trep);

groundPops=filterPops[w,fullbvecOn,vg];
excitedPops=filterPops[w,fullfluorvecOn,vg];
AppendTo[excitedPopsAll,excitedPops];

(* For debugging only; remove later *)
If[OptionValue[VerbosityLevel]>7,
Print[ListLinePlot[excitedPops,Mesh->All,AxesOrigin->{0.0,0.0},PlotRange->{{-1.0,1.0},All},Frame->True,FrameLabel->{"Frequency (GHz)","Excited State Population"}]];
];

(* Convert velocity classes, based on current time-dependent sol'n *)
If[OptionValue[VGMethod]===ConvertDM && i<iVGfreezeLoc,
(* This is a kludge... *)
NDitherLoc=1+Sign[FWHMbw/.params1];
speclist=MapThread[{#3,#1/(-Subtract@@#2)}&,{Total/@Partition[Total/@Partition[fullfluorvecOn  w,First[syst]],NDitherLoc],vg[[All,3]],vg[[All,1]]}];
lastvg=vg;
(* makeVg[speclist_,atdata1_,params_,opts:OptionsPattern[]]  *)
vg=makeVg[speclist,Last[syst],rfparams1,nfopts];
w=convertDM[w,lastvg,vg];
fullfluorvecOn=Flatten@Table[Evaluate@Take[fullfluorvecOn,Length[fullfluorvecOn]/Length[lastvg]],{Length[vg]}];
(* velocity group index of maximum excitation *)
excitedPopsHelp=filterPops[w,fullfluorvecOn,vg];
help1=Abs[vg[[All,1]]-0.5*NaturalWidth[2]/.Last[syst]];
indVgMaxEx=Position[help1,Min[help1]][[1,1]];
If[OptionValue[VerbosityLevel]>3,
Print["    indVgMaxEx: ",indVgMaxEx];];
, (* ELSE *)
excitedPopsHelp=excitedPops;
];

psiSpanAll=Join[psiSpanAll,psiSpan];
psiAvgTotal+=psiAvg;
errAll=Join[errAll,{err}];
psiAvgHist=Join[psiAvgHist,{psiAvg}];
(* Scale w to compensate for small normalization errors *)
(* w*=1.0/(popgr\[LeftDoubleBracket]-1,2\[RightDoubleBracket]+popex\[LeftDoubleBracket]-1,2\[RightDoubleBracket]);*)

(* Terminate iteration if two subsequent psi values are close to each other *)
terminationCriterion=(Abs[psiAvgLast/psiAvg-1.0]<TerminationThresh&&i>4);
(* Save data file: first 4 iterations, then once per CPU hour at the most, then last *)
fwriteCriterion=StringLength[fname]>0&&(i<5||terminationCriterion||(TimeUsed[]-tuLast)>3600.0||i==nperiods);

Global`res=If[terminationCriterion||fwriteCriterion||i==nperiods,
{psiAvg,psiAvgTotal,psiSpanAll,psiAvgHist,Iavg,popgr8All,popex16All,popgr8WeightedAll,{PhiVErrors->errAll,PsiIavg->psiIavg1,GroundPopulation->popgr,ExcitedPopulation->popex,FinalGroundPops->groundPops,FinalExcitedPops->excitedPops,ExcitedPops->excitedPopsAll,LastDM->{w,vg},RFParamsOut->{rfparams1,params1,opts}}}
,{}
];

If[fwriteCriterion,
(* Make sure "res" does not appear in the variable list of this Module (must not appear green in the following line), otherwise it cannot be loaded by Get[] into the global context! *)
Export[fname,Global`res];
tuLast=TimeUsed[];
If[OptionValue[VerbosityLevel]>2,Print[" ==> Saved (intermediate) results to ",fname];];
];
tu2=TimeUsed[];
If[OptionValue[VerbosityLevel]>1,Print["  Period ",i," done, CPU time {PsiMeso,MatrixExp,Rest}: ",{tu12-tu11,tu13-tu12,TimeUsed[]-tu13}," s, tot matrix prep time: ",tPrep,"s, psi avg: ",psiAvg,"\n"];];
(* Terminate iteration if two subsequent psi values are close to each other *)If[terminationCriterion,Print[" Last two psi values differ by only ",100.0*Abs[psiAvgLast/psiAvg-1.0],"%, terminating iteration now."];
Break[];];
psiAvgLast=psiAvg;
];(*For[i],main loop done*)
(* ============================ MAIN LOOP END =================================== *)

If[OptionValue[VerbosityLevel]>0,
Print["Completed, psi averaged over last period: ",psiAvg," ph/s/sr/atom/(W/m^2)"];Print["accumulated psi over all periods: ",psiAvgTotal ,", psi from avg cw irradiance: ",psiIavg1];
];
Global`res
];


(* ::Input::Initialization:: *)
End[]


(* ::Input::Initialization:: *)
Protect[Evaluate[$Context<>"*"]]


(* ::Input::Initialization:: *)
EndPackage[]
